@@ -1,3 +1,4 @@
+import os
 from itertools import product, combinations
 from functools import reduce
 from copy import deepcopy
@@ -69,6 +70,7 @@ class Section:
         Returns the specified subsection; we always need to run merge()
         on SectionParam first to avoid duplicates
         # TODO need special treatment for param sections with different interaction types (mostly dihedraltypes)
+        # TODO should we simply split them whenever intr type changes as we parse them?
         :param section_name:
         :return:
         """
@@ -79,6 +81,17 @@ class Section:
             raise RuntimeError("Error: subsection {} duplicated in {}".format(section_name, str(self)))
         return ssect[0]
 
+    def get_subsections(self, section_name):
+        """
+        Returns the list of specified subsections
+        :param section_name: str
+        :return: None
+        """
+        ssect = [s for s in self.subsections if s.header == section_name]
+        if len(ssect) == 0:
+            raise KeyError("Subsection {} not found, check your molecule topology!".format(section_name))
+        return ssect
+
 
 class SectionMol(Section):
     """
@@ -97,8 +110,7 @@ class SectionMol(Section):
 
     @property
     def atoms(self):
-        sub = self.get_subsection('atoms')
-        return [entry for entry in sub if isinstance(entry, gml.EntryAtom)]
+        return self.get_subsection('atoms').entries_atom
 
     @property
     def natoms(self):
@@ -108,11 +120,33 @@ class SectionMol(Section):
     def charge(self):
         return sum([a.charge for a in self.atoms])
 
-    def set_type(self, type, resname, atomname):
+    @property
+    def mass(self):
+        return sum([a.mass for a in self.atoms])
+
+    @property
+    def residues(self):
+        resid = None
+        reslist = []
+        for at in self.atoms:
+            if at.resid != resid:
+                resid = at.resid
+                reslist.append(f'{at.resname}-{at.resid}')
+        return reslist
+
+    def set_type(self, type_to_set, atomname, resname=None):
+        if resname is None:
+            resnames = list({a.resname for a in self.atoms})
+        else:
+            if isinstance(list, resname) or isinstance(tuple, resname):
+                resnames = resname
+            else:
+                resnames = [resname]
         for a in self.atoms:
-            if a.resname == resname and a.atomname == atomname:
-                a.type = type
-    
+            for rname in resnames:
+                if a.resname == rname and a.atomname == atomname:
+                    a.type = type_to_set
+
     def select_atoms(self, selection_string):
         """
         Returns atoms' indices according to the specified selection string
@@ -120,7 +154,7 @@ class SectionMol(Section):
         :return: list, 0-based indices of atoms compatible with the selection
         """
         sel = gml.SelectionParser(self)
-        return sel(selection_string)  # TODO enable getting atoms' properties
+        return sel(selection_string)
 
     def select_atom(self, selection_string):
         """
@@ -150,8 +184,8 @@ class SectionMol(Section):
     @property
     def is_alchemical(self):
         sect = self.get_subsection('atoms')
-        for ent in sect.entries:
-            if isinstance(ent, gml.EntryAtom) and ent.type_b is not None:
+        for ent in sect.entries_atom:
+            if ent.type_b is not None:
                 return True
         return False
 
@@ -161,8 +195,8 @@ class SectionMol(Section):
         if 'DTS' in resnames:
             selected = [x+1 for x in self.select_atoms('resname DTS')]
             sect = self.get_subsection('bonds')
-            for entry in sect:
-                if isinstance(entry, gml.EntryBonded) and all([x in selected for x in entry.atom_numbers]):
+            for entry in sect.entries_bonded:
+                if all([x in selected for x in entry.atom_numbers]):
                     entry.read_types()
                     if entry.atom_names == ("C5'", "O5'") or entry.atom_names == ("O5'", "C5'"):
                         if 'OS' in entry.types_state_a:
@@ -189,8 +223,8 @@ class SectionMol(Section):
         if 'DTD' in resnames or 'DTE' in resnames:
             selected = [x+1 for x in self.select_atoms('resname DTD DTE')]
             sect = self.get_subsection('bonds')
-            for entry in sect:
-                if isinstance(entry, gml.EntryBonded) and all([x in selected for x in entry.atom_numbers]):
+            for entry in sect.entries_bonded:
+                if all([x in selected for x in entry.atom_numbers]):
                     entry.read_types()
                     if entry.types_state_a == ("DUM_CT", "DUM_C2") \
                             and entry.atom_numbers[1] - entry.atom_numbers[0] > 30:
@@ -199,8 +233,8 @@ class SectionMol(Section):
         if 'DTX' in resnames or 'DTY' in resnames:
             selected = [x+1 for x in self.select_atoms('resname DTX DTY')]
             sect = self.get_subsection('bonds')
-            for entry in sect:
-                if isinstance(entry, gml.EntryBonded) and all([x in selected for x in entry.atom_numbers]):
+            for entry in sect.entries_bonded:
+                if all([x in selected for x in entry.atom_numbers]):
                     entry.read_types()
                     if entry.types_state_a == ("DUM_CT", "DUM_CT") \
                             and entry.atom_numbers[1] - entry.atom_numbers[0] > 30:
@@ -209,8 +243,8 @@ class SectionMol(Section):
 
     def nullify_bonded(self, atom1, atom2, subsection, stateB=True):
         subs = self.get_subsection(subsection)
-        for entry in subs:
-            if isinstance(entry, gml.EntryBonded) and atom1 in entry.atom_numbers and atom2 in entry.atom_numbers:
+        for entry in subs.entries_bonded:
+            if atom1 in entry.atom_numbers and atom2 in entry.atom_numbers:
                 entry.read_types()
                 if subsection == 'angles' and any(x.startswith('D') for x in entry.types_state_a):
                     continue
@@ -242,8 +276,8 @@ class SectionMol(Section):
         :return: None
         """
         subsection = self.get_subsection('atoms')
-        for entry_num, entry in enumerate(subsection):
-            if isinstance(entry, gml.EntryAtom) and entry.num >= startfrom:
+        for entry in subsection.entries_atom:
+            if entry.num >= startfrom:
                 entry.num += offset
 
     def _offset_params(self, offset, startfrom):
@@ -256,9 +290,8 @@ class SectionMol(Section):
         """
         for sub_name in [s.header for s in self.subsections if s.header != 'atoms']:
             subsection = self.get_subsection(sub_name)
-            for entry_num, entry in enumerate(subsection):
-                if isinstance(entry, gml.EntryBonded):
-                    entry.atom_numbers = tuple(n + (offset * (n >= startfrom)) for n in entry.atom_numbers)
+            for entry in subsection.entries_bonded:
+                entry.atom_numbers = tuple(n + (offset * (n >= startfrom)) for n in entry.atom_numbers)
 
     def gen_state_b(self, atomname=None, resname=None, resid=None, atomtype=None, new_type=None, new_charge=None,
                     new_mass=None):
@@ -275,15 +308,15 @@ class SectionMol(Section):
         :return: None
         """
         sub = self.get_subsection('atoms')
-        for entries in [entry for entry in sub if isinstance(entry, gml.EntryAtom)]:
-            criteria = all([(atomname is None or entries.atomname == atomname),
-                            (resname is None or entries.resname == resname),
-                            (resid is None or int(entries.resid) == int(resid)),
-                            (atomtype is None or entries.type == atomtype)])
+        for entry in sub.entries_atom:
+            criteria = all([(atomname is None or entry.atomname == atomname),
+                            (resname is None or entry.resname == resname),
+                            (resid is None or int(entry.resid) == int(resid)),
+                            (atomtype is None or entry.type == atomtype)])
             if criteria:
-                entries.type_b = new_type if new_type is not None else entries.type
-                entries.mass_b = new_mass if new_mass is not None else entries.mass
-                entries.charge_b = new_charge if new_charge is not None else entries.charge
+                entry.type_b = new_type if new_type is not None else entry.type
+                entry.mass_b = new_mass if new_mass is not None else entry.mass
+                entry.charge_b = new_charge if new_charge is not None else entry.charge
         self.update_dicts()
 
     def drop_state_a(self, remove_dummies=False, atomname=None, resname=None, resid=None, atomtype=None):
@@ -303,25 +336,25 @@ class SectionMol(Section):
         if atomname or resname or resid or atomtype:
             selected = set()
             sub = self.get_subsection('atoms')
-            for entries in [entry for entry in sub if isinstance(entry, gml.EntryAtom)]:
-                criteria = all([(atomname is None or entries.atomname == atomname),
-                                (resname is None or entries.resname == resname),
-                                (resid is None or int(entries.resid) == int(resid)),
-                                (atomtype is None or entries.type == atomtype)])
+            for entry in sub.entries_atom:
+                criteria = all([(atomname is None or entry.atomname == atomname),
+                                (resname is None or entry.resname == resname),
+                                (resid is None or int(entry.resid) == int(resid)),
+                                (atomtype is None or entry.type == atomtype)])
                 if criteria:
-                    selected.add(entries.num)
+                    selected.add(entry.num)
         else:
             selected = list(range(1, self.natoms+1))
         if remove_dummies:
             sub = self.get_subsection('atoms')
-            dummies = [entry for entry in sub if isinstance(entry, gml.EntryAtom) and entry.type_b and
-                       entry.type_b[0] == "D" and entry.num in selected]
+            dummies = [entry for entry in sub.entries_atom if entry.type_b and entry.type_b[0] == "D"
+                       and entry.num in selected]
             print(dummies)
             while dummies:
                 to_remove = dummies[-1]
                 self.del_atom(to_remove.num)
-                dummies = [entry for entry in sub if isinstance(entry, gml.EntryAtom) and entry.type_b and
-                           entry.type_b[0] == "D" and entry.num in selected]
+                dummies = [entry for entry in sub.entries_atom if entry.type_b and entry.type_b[0] == "D"
+                           and entry.num in selected]
         for sub in self.subsections:
             for entry in sub:
                 if (isinstance(entry, gml.EntryAtom) and entry.num in selected) \
@@ -349,13 +382,13 @@ class SectionMol(Section):
         if atomname or resname or resid or atomtype:
             selected = set()
             sub = self.get_subsection('atoms')
-            for entries in [entry for entry in sub if isinstance(entry, gml.EntryAtom)]:
-                criteria = all([(atomname is None or entries.atomname == atomname),
-                                (resname is None or entries.resname == resname),
-                                (resid is None or int(entries.resid) == int(resid)),
-                                (atomtype is None or entries.type == atomtype)])
+            for entry in sub.entries_atom:
+                criteria = all([(atomname is None or entry.atomname == atomname),
+                                (resname is None or entry.resname == resname),
+                                (resid is None or int(entry.resid) == int(resid)),
+                                (atomtype is None or entry.type == atomtype)])
                 if criteria:
-                    selected.add(entries.num)
+                    selected.add(entry.num)
         else:
             selected = list(range(1, self.natoms + 1))
         for sub in self.subsections:
@@ -388,13 +421,13 @@ class SectionMol(Section):
         if atomname or resname or resid or atomtype:
             selected = set()
             sub = self.get_subsection('atoms')
-            for entries in [entry for entry in sub if isinstance(entry, gml.EntryAtom)]:
-                criteria = all([(atomname is None or entries.atomname == atomname),
-                                (resname is None or entries.resname == resname),
-                                (resid is None or int(entries.resid) == int(resid)),
-                                (atomtype is None or entries.type == atomtype)])
+            for entry in sub.entries_atom:
+                criteria = all([(atomname is None or entry.atomname == atomname),
+                                (resname is None or entry.resname == resname),
+                                (resid is None or int(entry.resid) == int(resid)),
+                                (atomtype is None or entry.type == atomtype)])
                 if criteria:
-                    selected.add(entries.num)
+                    selected.add(entry.num)
         else:
             selected = list(range(1, self.natoms + 1))
         for sub in self.subsections:
@@ -409,13 +442,11 @@ class SectionMol(Section):
                         entry.types_state_b = None
         if remove_dummies:
             sub = self.get_subsection('atoms')
-            dummies = [entry for entry in sub if isinstance(entry, gml.EntryAtom) and entry.type[0] == "D"
-                       and entry.num in selected]
+            dummies = [entry for entry in sub.entries_atom if entry.type[0] == "D" and entry.num in selected]
             while dummies:
                 to_remove = dummies[-1]
                 self.del_atom(to_remove.num)
-                dummies = [entry for entry in sub if isinstance(entry, gml.EntryAtom) and entry.type[0] == "D"
-                           and entry.num in selected]
+                dummies = [entry for entry in sub.entries_atom if entry.type[0] == "D" and entry.num in selected]
         self.update_dicts()
     
     def add_atom(self, atom_number, atom_name, atom_type, charge=0.0, resid=None, resname=None, mass=None, prnt=True):
@@ -524,7 +555,7 @@ class SectionMol(Section):
                     raise RuntimeError("Could not match .top atom {} to a corresponding PDB atom".format(new_position))
                 old_loc = self._match_pdb_to_top(atom_number)[0]
                 new_loc = self._match_pdb_to_top(new_position)
-                atom = self.top.pdb.atoms.pop(old_loc-1)  # TODO check
+                atom = self.top.pdb.atoms.pop(old_loc-1)
                 self.top.pdb.atoms.insert(new_loc + 1, atom)
         subsect_atoms = self.get_subsection('atoms')
         atom_entry_list = [e for e in subsect_atoms.entries]
@@ -668,14 +699,15 @@ class SectionMol(Section):
         self.top.recalc_sys_params()
 
     def _merge_fields(self, other):
-        # TODO important: watch for POSRES
-        print('WARNING watch out for #ifdef POSRES keywords that might get misplaced')
+        self.top.print('WARNING watch out for #ifdef POSRES keywords that might get misplaced')
         for subs in ['atoms', 'bonds', 'angles', 'pairs', 'dihedrals', 'impropers', 'cmap', 'position_restraints']:
             # TODO merge all subsections
+            # TODO need a more consistent treatment of impropers
             try:
                 subsection_other = other.get_subsection(subs)
                 subsection_own = self.get_subsection(subs)
                 subsection_own.add_entries([entry for entry in subsection_other if entry])
+                self.top.print(f"Merging sections {subs} from two molecules")
             except KeyError:
                 pass
     
@@ -686,6 +718,7 @@ class SectionMol(Section):
         new_angles = self._generate_angles(other, atom_own, atom_other)
         new_pairs, new_dihedrals = self._generate_14(other, atom_own, atom_other)
         # TODO remove overlapping pairs between new_bond/new_angles and new_pairs for 4- and 5-membered rings
+        # or do we really need it?
         for sub, entries in zip(['bonds', 'pairs', 'angles', 'dihedrals'],
                                 [new_bond, new_pairs, new_angles, new_dihedrals]):
             subsection = self.get_subsection(sub)
@@ -777,14 +810,15 @@ class SectionMol(Section):
                           if (c, d) in self.bonds or (d, c) in self.bonds]
         return new_pairs, new_dihedrals
     
-    def add_ff_params(self, add_section='all'):
+    def add_ff_params(self, add_section='all', force_all=False):
         """
         Looks for FF parameters to be put for every bonded term in the topology,
         then adds them so that they can be explicitly seen/modified
         :param add_section: str, to which section should the FF params be added
+        :param force_all: bool, whether to overwrite existing parameters
         :return: None
         """
-        if add_section == 'all':  # TODO optionally add type/atomname labels in comment
+        if add_section == 'all':
             subsections_to_add = ['bonds', 'angles', 'dihedrals', 'impropers']
         else:
             subsections_to_add = [add_section]
@@ -795,11 +829,11 @@ class SectionMol(Section):
                 pass
             else:
                 for ssub in subsections:
-                    ssub.add_ff_params()
+                    ssub.add_ff_params(force_all)
 
     def find_used_ff_params(self, section='all'):
         used_params = []
-        if section == 'all':  # TODO optionally add type/atomname labels in comment
+        if section == 'all':
             subsections_to_add = ['bonds', 'angles', 'dihedrals', 'impropers']
         else:
             subsections_to_add = [section]
@@ -815,7 +849,7 @@ class SectionMol(Section):
 
     def find_missing_ff_params(self, add_section='all', fix_by_analogy=False, fix_B_from_A=False, fix_A_from_B=False,
                                once=False):
-        if add_section == 'all':  # TODO optionally add type/atomname labels in comment
+        if add_section == 'all':
             subsections_to_add = ['bonds', 'angles', 'dihedrals', 'impropers']
         else:
             subsections_to_add = [add_section]
@@ -844,13 +878,39 @@ class SectionMol(Section):
             else:
                 subsection.add_type_labels()
 
-    def mutate_protein_residue(self, resid, target, rtp=None, mutate_in_pdb=True):
-        alt_names = {('THR', 'OG'): 'OG1', ('THR', 'HG'): 'HG1', ('LEU', 'CD'): 'CD1', ('LEU', 'HD1'): 'HD11',
-                     ('LEU', 'HD2'): 'HD12', ('LEU', 'HD3'): 'HD13', ('VAL', 'CG'): 'CG1', ('VAL', 'HG1'): 'HG11',
-                     ('VAL', 'HG2'): 'HG12', ('VAL', 'HG3'): 'HG13',}
+    def add_posres(self, keyword, value):
+        try:
+            _ = self.get_subsection('position_restraints')
+        except KeyError:
+            try:
+                _ = value[0]
+            except TypeError:
+                value = 3 * [value]
+            content = ['[ position_restraints ]', f'#ifdef {keyword}', '; ai  funct  fcx    fcy    fcz']
+            for atom in self.atoms:
+                if len(atom.atomname) > 1:
+                    if not atom.atomname[0] == 'H' and not atom.atomname[1] == 'H':
+                        content.append(f"{atom.num:5}    1 {value[0]:5} {value[1]:5} {value[2]:5}")
+            content.append("#endif\n")
+            self.subsections.append(gml.SubsectionBonded(content, self))
+        else:
+            self.top.print(f"[ position_restraints ] already present in molecule {self.mol_name}, skipping")
+
+    def find_rtp(self, rtp):
+        """
+        Looks for aminoacids.rtp or merged.rtp in local files (*ff/*rtp or *rtp)
+        and in the Gromacs directory, then allows to interactively choose which one to use
+        :param rtp: str, path to the .rtp file (if applicable)
+        :return: str, path to the .rtp file found
+        """
         if rtp is None and not self.top.rtp:
+            found = glob(self.top.gromacs_dir + '/*ff/[am][em]*rtp') + glob(os.getcwd() + '/*ff/[am][em]*rtp') + \
+                    glob(os.getcwd() + '/[am][em]*rtp')
+            if not found:
+                raise RuntimeError("No .rtp files found locally or in default Gromacs dirs. Please set "
+                                   "rtp=/path/to/rtp/file")
             print("Found the following .rtp files:\n")
-            for n, i in enumerate(glob(self.top.gromacs_dir + '/*ff/[am][em]*rtp')):
+            for n, i in enumerate(found):
                 print('[', n + 1, '] ', i)
             rtpnum = input('\nPlease select one that contains the deserved charges and types:\n')
             try:
@@ -858,17 +918,32 @@ class SectionMol(Section):
             except ValueError:
                 raise RuntimeError('Not an integer: {}'.format(rtpnum))
             else:
-                rtp = glob(self.top.gromacs_dir + '/*ff/[am][em]*rtp')[rtpnum-1]
+                rtp = found[rtpnum-1]
         elif self.top.rtp:
             rtp = ''
+        return rtp
+
+    def mutate_protein_residue(self, resid, target, rtp=None, mutate_in_pdb=True):
+        """
+        Mutates an amino acid to a different one, optionally in the topology
+        and structure simultaneously
+        :param resid: int, number of the residue to be mutated
+        :param target: str, a single-letter code of the new residue to be introduced
+        :param rtp: str, path to the .rtp file that will be used to read atom properties
+        :param mutate_in_pdb: bool, whether to attempt modifying the same residue in the associated Pdb
+        :return: None
+        """
+        alt_names = {('THR', 'OG'): 'OG1', ('THR', 'HG'): 'HG1', ('LEU', 'CD'): 'CD1', ('LEU', 'HD1'): 'HD11',
+                     ('LEU', 'HD2'): 'HD12', ('LEU', 'HD3'): 'HD13', ('VAL', 'CG'): 'CG1', ('VAL', 'HG1'): 'HG11',
+                     ('VAL', 'HG2'): 'HG12', ('VAL', 'HG3'): 'HG13'}
+        rtp = self.find_rtp(rtp)
         orig = self.get_atom('resid {} and name CA'.format(resid))
         mutant = gml.ProteinMutant(orig.resname, target)
-        # TODO double-check [ pairs ]
         targ = mutant.target_3l
-        print("\n  Mutating residue {} (resid {}) into {}\n".format(orig.resname, resid, targ))
+        self.top.print("\n  Mutating residue {} (resid {}) into {}\n".format(orig.resname, resid, targ))
         atoms_add, hooks, _, _, extra_bonds, afters = mutant.atoms_to_add()
         atoms_remove = mutant.atoms_to_remove()
-        types, charges, impropers, improper_type = self.parse_rtp(rtp)
+        types, charges, dihedrals, impropers, improper_type = self.parse_rtp(rtp)
         # some residue-specific modifications here
         if targ == 'HIS':
             targ = 'HSD' if ('HSD', 'CA') in types.keys() else 'HID'
@@ -883,14 +958,14 @@ class SectionMol(Section):
         for at in atoms_remove:
             equivalents = {'OG': 'OG1', 'HG': 'HG1', 'HG1': 'HG11', 'HG2': 'HG12', 'HG3': 'HG13', 'CG': 'CG1',
                            'CD': 'CD1', 'HD': 'HD1', 'HD1': 'HD11', 'HD2': 'HD12', 'HD3': 'HD13'}
-            print("Removing atom {} from resid {} in topology".format(at, resid))
+            self.top.print("Removing atom {} from resid {} in topology".format(at, resid))
             try:
                 atnum = self.get_atom('resid {} and name {}'.format(resid, at)).num
             except RuntimeError:
                 atnum = self.get_atom('resid {} and name {}'.format(resid, equivalents[at])).num
             self.del_atom(atnum, del_in_pdb=False)
         for atom_add, hook, aft in zip(atoms_add, hooks, afters):
-            print("Adding atom {} to resid {} in topology".format(atom_add, resid))
+            self.top.print("Adding atom {} to resid {} in topology".format(atom_add, resid))
             # if there are ambiguities in naming (two or more options):
             if (targ, atom_add) in alt_names.keys():
                 atom_add = alt_names[(targ, atom_add)]
@@ -960,6 +1035,7 @@ class SectionMol(Section):
                         rid = resid
                         atx = at
                     numbers.append(atoms_sub.name_to_num[(rid, atx)])
+                # TODO if improper has extra params, add them here
                 new_str = '{:5d} {:5d} {:5d} {:5d} {:>5s}\n'.format(*numbers, improper_type)
                 impr_sub.add_entry(gml.EntryBonded(new_str, impr_sub),
                                    position=1+[n for n, e in enumerate(impr_sub) if isinstance(e, gml.EntryBonded)][-1])
@@ -987,25 +1063,38 @@ class SectionMol(Section):
             print("No .pdb file bound to the topology, use Top.add_pdb() to add one")
 
     def parse_rtp(self, rtp):
+        """
+        Reads an .rtp file to extract molecule definitions, separating them into
+        dictionaries for: types, charges, impropers, dihedrals, bondedtypes
+        :param rtp: str, path to the .rtp file
+        :return: tuple of dict, each containing atom name : relevant parameter matching
+        (ordered: types, charges, dihedrals, impropers, bondedtypes)
+        """
+        # TODO check against amber/ILDN
         if self.top.rtp:
-            return self.top.rtp['typedict'], self.top.rtp['chargedict'], self.top.rtp['impropers'], \
-                   self.top.rtp['bondedtypes']
+            return self.top.rtp['typedict'], self.top.rtp['chargedict'], self.top.rtp['dihedrals'], \
+                   self.top.rtp['impropers'], self.top.rtp['bondedtypes']
         chargedict, typedict = {}, {}
-        impropers = {}
+        impropers, dihedrals = {}, {}
         bondedtypes = 0
-        rtp_cont = [l for l in open(rtp) if not l.strip().startswith(';')]
+        rtp_cont = [line for line in open(rtp) if not line.strip().startswith(';')]
         resname = None
         reading_atoms = False
         reading_impropers = False
+        reading_dihedrals = False
         reading_bondedtypes = False
         for line in rtp_cont:
             if line.strip().startswith('[') and line.strip().split()[1] not in ['bondedtypes', 'atoms', 'bonds',
-                                                                                'impropers']:
+                                                                                'dihedrals', 'impropers']:
                 resname = line.strip().split()[1]
             if line.strip().startswith('[') and line.strip().split()[1] == 'atoms':
                 reading_atoms = True
             if line.strip().startswith('[') and line.strip().split()[1] != 'atoms':
                 reading_atoms = False
+            if line.strip().startswith('[') and line.strip().split()[1] == 'dihedrals':
+                reading_dihedrals = True
+            if line.strip().startswith('[') and line.strip().split()[1] != 'dihedrals':
+                reading_dihedrals = False
             if line.strip().startswith('[') and line.strip().split()[1] == 'impropers':
                 reading_impropers = True
             if line.strip().startswith('[') and line.strip().split()[1] != 'impropers':
@@ -1021,6 +1110,10 @@ class SectionMol(Section):
                 if resname not in impropers.keys():
                     impropers[resname] = []
                 impropers[resname].append(line.strip().split())
+            if len(line.strip().split()) > 3 and resname is not None and reading_dihedrals:
+                if resname not in dihedrals.keys():
+                    impropers[resname] = []
+                impropers[resname].append(line.strip().split())
             if len(line.strip().split()) > 7 and resname is None and reading_bondedtypes:
                 bondedtypes = line.strip().split()[3]
         # substitute CHARMM's HN for AMBER's H
@@ -1033,10 +1126,12 @@ class SectionMol(Section):
                 chargedict[(k[0], 'HG')] = chargedict[k]
         self.top.rtp['typedict'] = typedict
         self.top.rtp['chargedict'] = chargedict
+        self.top.rtp['dihedrals'] = dihedrals
         self.top.rtp['impropers'] = impropers
+        # TODO we need a method for explicitly setting just a single dihedral (should be easy)
         self.top.rtp['bondedtypes'] = bondedtypes
-        return self.top.rtp['typedict'], self.top.rtp['chargedict'], self.top.rtp['impropers'], \
-            self.top.rtp['bondedtypes']
+        return self.top.rtp['typedict'], self.top.rtp['chargedict'], self.top.rtp['dihedrals'], \
+            self.top.rtp['impropers'], self.top.rtp['bondedtypes']
 
     def update_dicts(self):
         self.get_subsection('atoms').get_dicts(force_update=True)
@@ -1102,7 +1197,7 @@ class SectionMol(Section):
         atoms_add, hooks = [basename.replace('C', 'H') + str(i) for i in range(3)], 3 * [orig_name]
         for n, atom_add_hook in enumerate(zip(atoms_add, hooks), 1):
             atom_add, hook = atom_add_hook
-            print("Adding atom {} to resid {} in the topology".format(atom_add, resid))
+            self.top.print("Adding atom {} to resid {} in the topology".format(atom_add, resid))
             hooksel = 'resid {} and name {}'.format(resid, orig_name)
             hnum = self.get_atom(hooksel).num
             atnum = hnum + n
@@ -1115,7 +1210,6 @@ class SectionMol(Section):
                 raise RuntimeError("Adding groups in PDB only supported for systems containing one molecule")
             bonds = self.list_bonds(returning=True)
             hook = [j for i in bonds for j in i if orig_name in i and orig_name != j][0]
-            print(hook)
             aligns = [j for i in bonds for j in i if hook in i and hook != j and orig_name != j]
             aftnr = self.select_atom('resid {} and name {}'.format(resid, orig_name))
             for n, aliat in enumerate(zip(aligns, atoms_add), 1):
@@ -1138,6 +1232,84 @@ class SectionMol(Section):
         self.add_dummy_def(dummy_type)
         self.gen_state_b(atomname=orig_name, resid=resid, new_type=dummy_type, new_charge=0, new_mass=1.008)
 
+    def solute_tempering(self, temperatures, selection=None):
+        self.top.explicit_defines()
+        for n, t in enumerate(temperatures):
+            self.top.print(f'generating topology for effective temperature of {t} K...')
+            mod = deepcopy(self.top)
+            mod.get_molecule(self.mol_name).scale_rest2_charges(temperatures[0]/t, selection)
+            mod.get_molecule(self.mol_name).scale_rest2_bonded(temperatures[0]/t, selection)
+            mod.save_top(self.top.fname.replace('.top', f'-rest{temperatures[0]/t:.3f}.top'))
+
+    def scale_rest2_bonded(self, gamma, selection=None):
+        # get a list of atomtypes to clone
+        sel = self.get_atoms(selection) if selection is not None else self.atoms
+        types = {at.type for at in sel}
+        for tp in types:
+            self.top.parameters.clone_type(tp)
+        for sub in ['atomtypes', 'bondtypes', 'angletypes', 'dihedraltypes', 'nonbond_params', 'pairtypes']:
+            try:
+                for subsect in self.top.parameters.get_subsections(sub):
+                    for ent in subsect.entries_param:
+                        if any([tp.startswith('y') for tp in ent.types]):
+                            ent.params[1] *= gamma
+            except KeyError:
+                pass
+        try:
+            for ent in self.top.parameters.get_subsection('cmaptypes').entries_param:
+                if any([tp.startswith('y') for tp in ent.types]):
+                    ent.params = [float(x)*gamma for x in ent.params]
+        except KeyError:
+            pass
+
+    def scale_rest2_charges(self, gamma, selection=None):
+        sel = self.get_atoms(selection) if selection is not None else self.atoms
+        # scaling charges
+        for a in self.atoms:
+            if a in sel:
+                a.charge = round(a.charge * gamma**0.5, 4)
+                a.type = 'y' + a.type
+
+    def alchemical_proton(self, resid, rtp=None, b_is_protonated=False):
+        """
+        Creates an alchemical residue (starting from ASP or GLU) where the B-state
+        corresponds to a protonated variant of that residue
+        :param resid: int, number of the residue to modify
+        :param rtp: str, path to the aminoacids.rtp or merged.rtp file (optional)
+        :param b_is_protonated: bool, whether to make the B-state protonated
+        :return: None
+        """
+        atoms = self.get_atoms(f'resid {resid}')
+        resname = atoms[0].resname
+        if resname not in ['ASP', 'GLU']:
+            raise RuntimeError("So far only available for residues ASP and GLU")
+        mut_dict = {'ASP': 'B', 'GLU': 'J'}
+        self.mutate_protein_residue(resid, mut_dict[resname])
+        atoms = self.get_atoms(f'resid {resid}')
+        rtp = self.find_rtp(rtp)
+        types, charges, dihedrals, impropers, improper_type = self.parse_rtp(rtp)
+        for atom in atoms:
+            try:
+                atom.type_b = types[(resname, atom.atomname)]
+                atom.charge_b = charges[(resname, atom.atomname)]
+                atom.mass_b = atom.mass
+            except KeyError:
+                atom.type_b = 'DH'
+                atom.charge_b = 0.0
+                atom.mass_b = 1.008
+        if 'DH' not in self.top.defined_atomtypes:
+            self.top.parameters.get_subsection('atomtypes').add_entry('DH  0  0.0  0.0  A   0.0  0.0')
+        if b_is_protonated:
+            self.swap_states(resid=resid)
+        self.update_dicts()
+        self.add_ff_params()
+
+    def recalc_qtot(self):
+        charge = 0
+        for atom in self.atoms:
+            charge += atom.charge
+            atom.comment = f'qtot {charge:.3:f}'
+
 
 class SectionParam(Section):
     """
@@ -1161,6 +1333,8 @@ class SectionParam(Section):
         """
         subsection_labels = [sub.label for sub in self.subsections]
         duplicated_subsections = list({label for label in subsection_labels if subsection_labels.count(label) > 1})
+        if duplicated_subsections:
+            self.top.print(f"Merging sections {', '.join(duplicated_subsections)} together")
         for sub in duplicated_subsections:
             subsections_to_merge = [s for s in self.subsections if s.label == sub]
             merged_subsection = reduce(lambda x, y: x+y, subsections_to_merge)
@@ -1187,7 +1361,7 @@ class SectionParam(Section):
 
     def find_used_ff_params(self, section='all'):
         used_params = []
-        if section == 'all':  # TODO optionally add type/atomname labels in comment
+        if section == 'all':
             subsections_to_add = ['atomtypes', 'pairtypes', 'nonbonded_params', 'constrainttypes']
         else:
             subsections_to_add = [section]
@@ -1211,7 +1385,9 @@ class SectionParam(Section):
         :return: None
         """
         new_atomtype = prefix + atomtype if new_type is None else new_type
-        # TODO check first if new_atomtype overlaps with existing ones
+        existing_types = {e.types[0] for e in self.top.parameters.get_subsection('atomtypes').entries_param}
+        if new_atomtype in existing_types:
+            raise RuntimeError(f"Type {new_atomtype} already exists")
         for sub in self.subsections:
             to_add = []
             for ent in sub:
