@@ -1,5 +1,4 @@
 import gromologist as gml
-from copy import deepcopy
 
 
 class Subsection:
@@ -152,6 +151,15 @@ class SubsectionBonded(Subsection):
     
     def __repr__(self):
         return "Subsection {} with interaction type {}".format(self.header, self.prmtype)
+
+    @property
+    def entries_bonded(self):
+        """
+        Lists all entries that are of EntryBonded type (i.e. excluding comments,
+        headers, empty lines etc)
+        :return: list of EntryBonded instances
+        """
+        return [e for e in self.entries if isinstance(e, gml.EntryBonded)]
     
     def sort(self):
         """
@@ -173,33 +181,63 @@ class SubsectionBonded(Subsection):
         return val
 
     def explicit_defines(self):
-        for entry in self.entries:
-            if isinstance(entry, gml.EntryBonded):
-                entry.explicit_defines()
+        """
+        Substitutes predefined fields with their corresponding values
+        as specified by the #DEFINE preprocessor commands
+        :return: None
+        """
+        for entry in self.entries_bonded:
+            entry.explicit_defines()
     
-    def add_ff_params(self):
+    def add_ff_params(self, force_all=False):
+        """
+        Adds explicit values of FF parameters in sections 'bonds', 'angles', 'dihedrals'
+        :param force_all: bool, whether to add params from scratch if some are assigned already
+        :return: None
+        """
         matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes',
                      'impropers': 'dihedraltypes'}
-        subsect_params = [sub for sub in self.section.top.parameters.subsections if sub.header == matchings[self.header]]
+        subsect_params = [sub for sub in self.section.top.parameters.subsections
+                          if sub.header == matchings[self.header]]
         self.bkp_entries = self.entries[:]  # we can't change what we're iterating over, so we modify the copy
-        for entry in self.entries:
-            if isinstance(entry, gml.EntryBonded) and not entry.params_state_a:
-                self._add_ff_params_to_entry(entry, subsect_params)
+        for entry in self.entries_bonded:
+            # let's omit the ones that already have params assigned:
+            if entry.params_state_a:
+                if not force_all:
+                    continue
+                else:
+                    entry.params_state_a = []
+                    entry.params_state_b = []
+                    entry.params_state_a_entry = []
+                    entry.params_state_b_entry = []
+            self._add_ff_params_to_entry(entry, subsect_params)
         self.entries = self.bkp_entries[:]  # now restore the modified copy
 
     def find_used_ff_params(self):
+        """
+        Looks up a list of FF parameters that are actually used in the molecular
+        system at hand
+        :return: list of str, labels of used parameters
+        """
         used_parm_entries = []
         matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes',
                      'impropers': 'dihedraltypes'}
         subsect_params = [sub for sub in self.section.top.parameters.subsections if
                           sub.header == matchings[self.header]]
-        for entry in self.entries:
-            if isinstance(entry, gml.EntryBonded) and not entry.params_state_a:
+        for entry in self.entries_bonded:
+            if not entry.params_state_a:
                 used_parm_entries.extend(self._find_used_ff_params(entry, subsect_params))
         return used_parm_entries
 
     @staticmethod
     def _find_used_ff_params(entry, subsect_params):
+        """
+        Identifies the FF parameter(s) that is used for the particular
+        bonded interaction entry
+        :param entry: a BondedEntry instance, an entry we're finding the parameter for
+        :param subsect_params: a SubsectionParam instance containing candidate parameters for the entry
+        :return: list of str, labels of used parameters
+        """
         entries = []
         int_type = entry.interaction_type
         entry.read_types()
@@ -207,7 +245,7 @@ class SubsectionBonded(Subsection):
             wildcard_present = []
             non_wildcard_present = []
             for subsections in subsect_params:
-                for parm_entry in [e for e in subsections if isinstance(e, gml.EntryParam)]:
+                for parm_entry in [e for e in subsections.entries_param]:
                     if parm_entry.match(types, int_type):
                         is_wildcard = 'X' in parm_entry.types
                         if not wildcard_present and not is_wildcard:
@@ -227,13 +265,21 @@ class SubsectionBonded(Subsection):
         return entries
 
     def find_missing_ff_params(self, fix_by_analogy=False, fix_B_from_A=False, fix_A_from_B=False, once=False):
+        """
+        Identifies FF parameters that cannot by matched to the existing set of
+        bonded interactions, allowing to find & target these for fixing
+        :param fix_by_analogy: dict, tells how to choose new parameters from existing equivalent types
+        :param fix_B_from_A: bool, chooses missing parameters assuming B-state should have similar types to A-state
+        :param fix_A_from_B: bool, chooses missing parameters assuming A-state should have similar types to B-state
+        :param once: bool, only shows the missing parameter once if repeated
+        :return: None
+        """
         matchings = {'bonds': 'bondtypes', 'angles': 'angletypes', 'dihedrals': 'dihedraltypes',
                      'impropers': 'dihedraltypes'}
         subsect_params = [sub for sub in self.section.top.parameters.subsections if
                           sub.header == matchings[self.header]]
-        for entry in self.entries:
-            if isinstance(entry, gml.EntryBonded):
-                self._find_missing_ff_params(entry, subsect_params, fix_by_analogy, fix_B_from_A, fix_A_from_B, once)
+        for entry in self.entries_bonded:
+            self._find_missing_ff_params(entry, subsect_params, fix_by_analogy, fix_B_from_A, fix_A_from_B, once)
 
     def _find_missing_ff_params(self, entry, subsect_params, fix_by_analogy, fix_B_from_A, fix_A_from_B, once):
         if (entry.params_state_a and entry.params_state_b) or (entry.params_state_a and not entry.types_state_b):
@@ -242,7 +288,7 @@ class SubsectionBonded(Subsection):
         entry.read_types()
         found_a, found_b = False, False
         for subsections in subsect_params:
-            for parm_entry in [e for e in subsections if isinstance(e, gml.EntryParam)]:
+            for parm_entry in subsections.entries_param:
                 if parm_entry.match(entry.types_state_a, int_type):
                     found_a = True
         if not found_a and not entry.params_state_a:
@@ -257,7 +303,7 @@ class SubsectionBonded(Subsection):
                     entry.params_state_a = candid
         if entry.types_state_b and not entry.params_state_b:
             for subsections in subsect_params:
-                for parm_entry in [e for e in subsections if isinstance(e, gml.EntryParam)]:
+                for parm_entry in subsections.entries_param:
                     if parm_entry.match(entry.types_state_b, int_type):
                         found_b = True
             if not found_b and not entry.params_state_b:
@@ -291,7 +337,7 @@ class SubsectionBonded(Subsection):
         else:
             types = other_typelist
         for subsection in subsect_params:
-            for parm_entry in [e for e in subsection if isinstance(e, gml.EntryParam)]:
+            for parm_entry in subsection.entries_param:
                 if parm_entry.match(types, int_type):
                     if from_wildtype is None:
                         if 'X' in parm_entry.types:
@@ -300,7 +346,8 @@ class SubsectionBonded(Subsection):
                             from_wildtype = False
                         new_params.append(parm_entry.params)
                         print(f'Fixing by analogy, using entry {str(parm_entry).strip()}')
-                    elif (from_wildtype and 'X' in parm_entry.types) or (not from_wildtype and 'X' not in parm_entry.types):
+                    elif (from_wildtype and 'X' in parm_entry.types) or (not from_wildtype
+                                                                         and 'X' not in parm_entry.types):
                         new_params.append(parm_entry.params)
                         print(f'Fixing by analogy, using entry {str(parm_entry).strip()}')
                     if len(typelist) < 4:
@@ -321,12 +368,12 @@ class SubsectionBonded(Subsection):
         int_type = entry.interaction_type
         entry.read_types()
         for types, params, parmentry in zip([entry.types_state_a, entry.types_state_b],
-                                        [entry.params_state_a, entry.params_state_b],
-                                        [entry.params_state_a_entry, entry.params_state_b_entry]):
+                                            [entry.params_state_a, entry.params_state_b],
+                                            [entry.params_state_a_entry, entry.params_state_b_entry]):
             wildcard_present = []
             non_wildcard_present = []
             for subsections in subsect_params:
-                for parm_entry in [e for e in subsections if isinstance(e, gml.EntryParam)]:
+                for parm_entry in subsections.entries_param:
                     if parm_entry.match(types, int_type):
                         is_wildcard = 'X' in parm_entry.types
                         if not wildcard_present and not is_wildcard:
@@ -405,15 +452,13 @@ class SubsectionBonded(Subsection):
         while Amber uses angletype '1' (simple harmonic)
         :return: str, interaction type
         """
-        for entry in self:
-            if isinstance(entry, gml.EntryBonded):
-                return entry.interaction_type
+        for entry in self.entries_bonded:
+            return entry.interaction_type
         return '0'
 
     def add_type_labels(self):
-        for entry in self.entries:
-            if isinstance(entry, gml.EntryBonded) and not entry.params_state_a:
-                self._add_type_label(entry)
+        for entry in self.entries_bonded:
+            self._add_type_label(entry)
 
     @staticmethod
     def _add_type_label(entry):
@@ -462,6 +507,10 @@ class SubsectionParam(Subsection):
                                                                                                 other.header))
         return SubsectionParam(["[ {} ]\n".format(self.header)] + self.entries + other.entries, self.section)
 
+    @property
+    def entries_param(self):
+        return [e for e in self.entries if isinstance(e, gml.EntryParam)]
+
     def sort(self):
         """
         In case we want to sort entries after some are added at the end of the section
@@ -474,15 +523,12 @@ class SubsectionParam(Subsection):
 
     def find_used_ff_params(self):
         used_parm_entries = []
-        used_atomtypes_a = {a.type for mol in self.section.top.molecules for a in mol.atoms
-                            if isinstance(a, gml.EntryAtom)}
-        used_atomtypes_b = {a.type_b for mol in self.section.top.molecules for a in mol.atoms
-                            if isinstance(a, gml.EntryAtom) and a.type_b is not None}
+        used_atomtypes_a = {a.type for mol in self.section.top.molecules for a in mol.atoms}
+        used_atomtypes_b = {a.type_b for mol in self.section.top.molecules for a in mol.atoms if a.type_b is not None}
         used_atomtypes = used_atomtypes_a.union(used_atomtypes_b)
-        for entry in self.entries:
-            if isinstance(entry, gml.EntryParam):
-                if all(tp in used_atomtypes for tp in entry.types):
-                    used_parm_entries.append(entry.identifier)
+        for entry in self.entries_param:
+            if all(tp in used_atomtypes for tp in entry.types):
+                used_parm_entries.append(entry.identifier)
         return used_parm_entries
 
     def _sorting_fn(self, entry):
@@ -516,25 +562,24 @@ class SubsectionParam(Subsection):
         return '0'
 
     def get_opt_dih(self, types=False):
-        dopts = [entry for entry in self.entries if isinstance(entry, gml.EntryParam) and 'DIHOPT' in entry.comment]
+        dopts = [entry for entry in self.entries_param if 'DIHOPT' in entry.comment]
         if not types:
             return [e.params[x] for e in dopts for x in [0, 1]]
         else:
             return [(*e.types, e.params[2]) for e in dopts]
 
     def get_opt_dih_indices(self):
-        dopts = [entry for entry in self.entries if isinstance(entry, gml.EntryParam) and 'DIHOPT' in entry.comment]
+        dopts = [entry for entry in self.entries_param if 'DIHOPT' in entry.comment]
         self.section.top.add_ff_params()
         indices = set()
-        for i in dopts:  # TODO so far only works for one molecule
-            for j in self.section.top.molecules[0].get_subsection('dihedrals'):
-                if isinstance(j, gml.EntryBonded):
-                    if i in j.params_state_a_entry:
-                        indices.add(j.atom_numbers)
+        for i in dopts:  # TODO so far only works for one (first) molecule
+            for j in self.section.top.molecules[0].get_subsection('dihedrals').entries_bonded:
+                if i in j.params_state_a_entry:
+                    indices.add(j.atom_numbers)
         return list(indices)
 
     def set_opt_dih(self, values):
-        dopts = [entry for entry in self.entries if isinstance(entry, gml.EntryParam) and 'DIHOPT' in entry.comment]
+        dopts = [entry for entry in self.entries_param if 'DIHOPT' in entry.comment]
         for e, ang, k in zip(dopts, values[::2], values[1::2]):
             e.params[0] = ang
             e.params[1] = k
@@ -591,6 +636,10 @@ class SubsectionAtom(Subsection):
         self.fstring = "{:>6}{:>11}{:>7}{:>7}{:>7}{:>7}{:>11}{:>11}   ; " + '\n'
         self.name_to_num, self.num_to_name, self.num_to_type, self.num_to_type_b = None, None, None, None
 
+    @property
+    def entries_atom(self):
+        return [e for e in self.entries if isinstance(e, gml.EntryAtom)]
+
     def get_dicts(self, force_update=False):
         """
         dicts are not always needed and are costly to calculate,
@@ -609,12 +658,11 @@ class SubsectionAtom(Subsection):
         molname:(num:type) bindings
         """
         name_to_num, num_to_name, num_to_type, num_to_type_b = {}, {}, {}, {}
-        for entry in self:
-            if isinstance(entry, gml.EntryAtom):
-                name_to_num[(entry.resid, entry.atomname)] = entry.num
-                num_to_name[entry.num] = entry.atomname
-                num_to_type[entry.num] = entry.type
-                num_to_type_b[entry.num] = entry.type_b if entry.type_b is not None else entry.type
+        for entry in self.entries_atom:
+            name_to_num[(entry.resid, entry.atomname)] = entry.num
+            num_to_name[entry.num] = entry.atomname
+            num_to_type[entry.num] = entry.type
+            num_to_type_b[entry.num] = entry.type_b if entry.type_b is not None else entry.type
         return name_to_num, num_to_name, num_to_type, num_to_type_b
 
     
