@@ -1,13 +1,14 @@
 import os
 import datetime
+from copy import deepcopy
 
 import gromologist as gml
 from collections import OrderedDict
 
 
 class Top:
-    def __init__(self, filename, gmx_dir=None, pdb=None, ignore_ifdef=False, define=None, ifdef=None, keep_all=False,
-                 suppress=False):
+    def __init__(self, filename, gmx_dir=None, gmx_exe=None, pdb=None, ignore_ifdef=False, define=None, ifdef=None,
+                 keep_all=True, suppress=False):
         """
         A class to represent and contain the Gromacs topology file and provide
         tools for editing topology elements
@@ -19,10 +20,9 @@ class Top:
         """
         # TODO maybe allow for construction of a blank top with a possibility to read data later?
         self.suppress = suppress
-        if not gmx_dir:
-            self.gromacs_dir = self.find_gmx_dir()
-        else:
-            self.gromacs_dir = gmx_dir
+        self.gromacs_dir, self.gmx_exe = self.find_gmx_dir()
+        self.gromacs_dir = gmx_dir if not self.gromacs_dir else self.gromacs_dir
+        self.gmx_exe = gmx_exe if not self.gmx_exe else self.gmx_exe
         self.pdb = None
         self.rtp = {}
         self.pdb = None if pdb is None else gml.Pdb(pdb, top=self)
@@ -34,20 +34,23 @@ class Top:
         self.defines = {}
         if define is not None:
             self.defines.update(define)
+        self._preprocess_conditional_includes(ifdef)
         self._include_all(ignore_ifdef)
         if not keep_all:
             self.resolve_ifdefs([] if ifdef is None else ifdef)
-            self.print("Trying to keep all conditional (#ifdef/#endif) sections, this might lead to issues if a section"
-                       " is defined more than once with different conditions - check your topology afterwards!")
         else:
-            pass
+            self.print("Keeping all conditional (#ifdef/#endif) sections, this might lead to issues if sections "
+                       "are being merged or moved around - look out for messages & check your topology afterwards!")
         self.sections = []
         self.header = []
         self._parse_sections()
+
+    @property
+    def system(self):
         if self.top.endswith('top'):
-            self._read_system_properties()
+            return {}
         else:
-            self.system = {}
+            return self.read_system_properties()
 
     def __repr__(self):
         return "Topology with {} atoms and total charge {:.3f}".format(self.natoms, self.charge)
@@ -76,15 +79,14 @@ class Top:
             gmx = os.popen('which gmx_mpi 2> /dev/null').read().strip()
         if not gmx:
             gmx = os.popen('which gmx_d 2> /dev/null').read().strip()
-        if not gmx:
-            gmx = os.popen('which grompp 2> /dev/null').read().strip()
         if gmx:
-            gmx = '/'.join(gmx.split('/')[:-2]) + '/share/gromacs/top'
-            self.print('Gromacs files found in directory {}'.format(gmx))
-            return gmx
+            gmx_dir = '/'.join(gmx.split('/')[:-2]) + '/share/gromacs/top'
+            self.print('Gromacs files found in directory {}'.format(gmx_dir))
+            return gmx_dir, gmx
         else:
-            self.print('No working Gromacs compilation found, assuming all file dependencies are referred to locally')
-            return ""
+            self.print('No working Gromacs compilation found, assuming all file dependencies are referred to locally; '
+                       'to change this, make the gmx executable visible in $PATH or specify gmx_dir for the Topology')
+            return False, False
 
     @property
     def molecules(self):
@@ -108,6 +110,10 @@ class Top:
                     atomlist.append(a)
         return atomlist
 
+    @property
+    def defined_atomtypes(self):
+        return {ent.types[0] for ent in self.parameters.get_subsection('atomtypes').entries_param}
+
     def list_molecules(self):
         """
         Prints out a list of molecules contained in the System
@@ -152,11 +158,16 @@ class Top:
         :param fix_by_analogy: dict, if set, will attempt to use params by analogy, matching key types to value types
         :param fix_B_from_A: bool, will assign params for state B from state A
         :param fix_A_from_B: bool, will assign params for state A from state B
-        :param once: bool, will only print a given missing term once
+        :param once: bool, will only print a given missing term once per molecule
         :return: None
         """
         for mol in self.molecules:
-            mol.find_missing_ff_params(section, fix_by_analogy, fix_B_from_A, fix_A_from_B)
+            mol.find_missing_ff_params(section, fix_by_analogy, fix_B_from_A, fix_A_from_B, once=once)
+
+    def add_posres(self, keyword='POSRES', value=1000):
+        for mol in self.molecules:
+            if len(mol.atoms) > 3:
+                mol.add_posres(keyword, value)
 
     def add_params_file(self, paramfile):
         prmtop = Top._from_text('#include {}\n'.format(paramfile))
@@ -190,6 +201,35 @@ class Top:
             lines = [i for i in range(len(self._contents)) if self._contents[i].startswith("#include")
                      and i not in ignore_lines]
 
+    def _preprocess_conditional_includes(self, ifdef):
+        """
+        Because of the implementation of CHARMM36m/CHARMM36 in gmx,
+        the two variants are chosen depending on a preprocessing conditional
+        so we need to pre-treat the topology to account for that
+        -- I know it's extremely ugly but all other options require too much from the user --
+        :param ifdef: list of str, defined keywords
+        :return:
+        """
+        start_final = []
+        flag = False
+        for n, line in enumerate(self._contents):
+            if len(line.split()) > 2 and line.split()[0] == '#ifdef' and line.split()[1] == "USE_OLD_C36":
+                start_final.append(n)
+            elif len(start_final) == 1 and line.strip().startswith('#include'):
+                flag = True
+            elif len(start_final) == 1 and len(line.split()) > 1 and line.split()[0] == '#endif' and flag:
+                start_final.append(n)
+        if len(start_final) == 2:
+            if "USE_OLD_C36" in ifdef:
+                incl = '#include "old_c36_cmap.itp"\n'
+            else:
+                self.top.print('Will use (newer) CMAP parameters for CHARMM36m. To use CHARMM36, '
+                               'specify ifdef=["USE_OLD_C36"] (if your FF version supports that)')
+                incl = '#include "cmap.itp"\n'
+            for lnum in range(start_final[0], start_final[1]+1):
+                self._contents.pop(lnum)
+            self._contents.insert(start_final[0], incl)
+
     def _find_ifdef_lines(self):
         """
         Finds #ifdef/#endif blocks in the topology if user
@@ -216,6 +256,10 @@ class Top:
             keyword = None
             header_in_ifdef = False
             for n, line in enumerate(self._contents):
+                if line.strip().startswith('[') and counter > 0 and keyword != 'USE_OLD_C36':
+                    header = line.strip().strip('[]').strip()
+                    raise RuntimeError(f"Found a [ {header} ] section header within a conditional block, this can lead"
+                                       f"to issues if sections are rearranged. Please fix this first.")
                 if line.startswith('#define') and len(line.strip().split()) == 2:
                     ifdefs.append(line.strip().split()[1])
                 if line.strip().startswith("#ifdef") or line.strip().startswith("#ifndef"):
@@ -244,10 +288,11 @@ class Top:
                         counter -= 1
                         if header_in_ifdef:
                             if remove_lines and keyword not in ifdefs:
-                                self.print("Part of the #ifdef {} section ({} lines) will be dropped as specified because "
-                                          "the keyword was not defined. To prevent it, add 'ifdef=['{}']' as an argument "
-                                          "when loading the topology to define a keyword, or set 'keep_all=True' (might "
-                                          "occasionally produce issues).".format(keyword, len(remove_lines), keyword))
+                                self.print("Part of the #ifdef {} section ({} lines) will be dropped as specified "
+                                           "because the keyword was not defined. To prevent it, add 'ifdef=['{}']' "
+                                           "as an argument when loading the topology to (unconditionally) keep "
+                                           "specific sections, or set 'keep_all=True' to preserve the original (might "
+                                           "occasionally produce issues).".format(keyword, len(remove_lines), keyword))
                             elif remove_lines and keyword in ifdefs:
                                 self.print("Part of the #ifdef {} section ({} lines) will be dropped "
                                            "as specified.".format(keyword, len(remove_lines), keyword))
@@ -290,7 +335,9 @@ class Top:
             filename = filename.strip()[2:]
         pref = '/'.join(filename.split('/')[:-1])
         suff = filename.split('/')[-1]
-        if os.path.isfile(self.dir.rstrip(os.sep) + os.sep + pref + os.sep + suff):
+        if filename.startswith('/') and os.path.isfile(filename):
+            return filename, pref
+        elif os.path.isfile(self.dir.rstrip(os.sep) + os.sep + pref + os.sep + suff):
             return self.dir.rstrip(os.sep) + os.sep + pref + os.sep + suff, pref
         elif os.path.isfile(self.gromacs_dir.rstrip(os.sep) + os.sep + pref + os.sep + suff):
             return self.gromacs_dir.rstrip(os.sep) + os.sep + pref + os.sep + suff, pref
@@ -351,7 +398,7 @@ class Top:
         else:
             return gml.Section(content, self)
 
-    def _read_system_properties(self):
+    def read_system_properties(self):
         """
         Reads in system composition based on the [ molecules ] section
         :return: None
@@ -363,12 +410,11 @@ class Top:
             raise RuntimeError("Multiple 'molecules' subsection found in the topology, this is not allowed")
         elif len(system_subsection) == 0:
             self.print("Section 'molecules' not present in the topology, assuming this is an isolated .itp")
-            self.system = None
-            return
+            return None
         for e in system_subsection[0]:
             if e.content:
                 molecules[e.content[0]] = int(e.content[1])
-        self.system = molecules
+        return molecules
 
     @property
     def charge(self):
@@ -378,15 +424,6 @@ class Top:
     def natoms(self):
         return sum([self.system[mol] * self.get_molecule(mol).natoms for mol in self.system.keys()])
 
-    def recalc_sys_params(self):
-        """
-        Wrapper around read_system_properties that also
-        recalculates relevant molecule properties after changes
-        :return: None
-        """
-        sub_mol = [sub for sect in self.sections for sub in sect.subsections if isinstance(sub, gml.SubsectionAtom)]
-        self._read_system_properties()
-
     def explicit_defines(self):
         """
         Changes pre-defined keywords in parameter sets
@@ -476,6 +513,20 @@ class Top:
                     a.atomname = a.atomname.replace('DH', 'Hx').replace('DO', 'Ox').replace('DN', 'Nx').\
                         replace('DC', 'Cx').replace('DS', 'Sx')
 
+    def recalculate_qtot(self):
+        for mol in self.molecules:
+            mol.recalc_qtot()
+
+    def solute_tempering(self, temperatures, molecules):
+        self.explicit_defines()
+        for n, t in enumerate(temperatures):
+            self.print(f'generating topology for effective temperature of {t} K...')
+            mod = deepcopy(self)
+            mod.molecules[0].scale_rest2_bonded(temperatures[0] / t)
+            for i in molecules:
+                mod.molecules[i].scale_rest2_charges(temperatures[0]/t)
+            mod.save_top(self.fname.replace('.top', f'-rest{temperatures[0]/t:.3f}.top'))
+
     @staticmethod
     def _write_section(outfile, section):
         for subsection in section.subsections:
